from flask import _app_ctx_stack as stack
import os
from pymongo import MongoClient
import redis as redispy
from werkzeug.local import LocalProxy


def _connect_mongo_db():
    if os.environ.get('MONGOHQ_URL'):
        c = MongoClient(os.environ['MONGOHQ_URL'])
        db = c.get_default_database()
    else:
        c = MongoClient()
        db = c.test

    return db


def _connect_redis():
    redis_url = os.getenv('REDISTOGO_URL', 'redis://localhost:6379')
    return redispy.from_url(redis_url)


def _get_mongo_db():
    ctx = stack.top
    if ctx is not None:
        if not hasattr(ctx, 'mongo_db'):
            ctx.mongo_db = _connect_mongo_db()
        return ctx.mongo_db


def _get_redis():
    ctx = stack.top
    if ctx is not None:
        if not hasattr(ctx, 'redis'):
            ctx.redis = _connect_redis()
        return ctx.redis

redis = LocalProxy(_get_redis)
mongo = LocalProxy(_get_mongo_db)


import json
import logging

from ac_flask.hipchat.events import events
from ac_flask.hipchat.db import mongo, redis
from .tenant import Tenant
from flask import request
import requests


_log = logging.getLogger(__name__)


def _invalid_install(msg):
    _log.error("Installation failed: %s" % msg)
    return msg, 400


def init(addon, allow_global, allow_room, send_events=True, db_name='clients', require_group_id=False):

    # noinspection PyUnusedLocal
    @addon.app.route('/addon/installable', methods=['POST'])
    def on_install():
        clients = mongo[db_name]
        data = json.loads(request.data)
        if not data.get('roomId', None) and not allow_global:
            return _invalid_install("This add-on can only be installed in individual rooms.  Please visit the " +
                                    "'Add-ons' link in a room's administration area and install from there.")

        if data.get('roomId', None) and not allow_room:
            return _invalid_install("This add-on cannot be installed in an individual room.  Please visit the " +
                                    "'Add-ons' tab in the 'Group Admin' area and install from there.")

        _log.info("Retrieving capabilities doc at %s" % data['capabilitiesUrl'])
        capdoc = requests.get(data['capabilitiesUrl'], timeout=10).json()

        if capdoc['links'].get('self', None) != data['capabilitiesUrl']:
            return _invalid_install("The capabilities URL %s doesn't match the resource's self link %s" %
                                    (data['capabilitiesUrl'], capdoc['links'].get('self', None)))

        client = Tenant(data['oauthId'], data['oauthSecret'], room_id=data.get('roomId', None), capdoc=capdoc)

        try:
            session = client.get_token(redis, token_only=False,
                                       scopes=addon.descriptor['capabilities']['hipchatApiConsumer']['scopes'])
        except Exception as e:
            _log.warn("Error validating installation by receiving token: %s" % e)
            return _invalid_install("Unable to retrieve token using the new OAuth information")

        _log.info("session: %s" % json.dumps(session))
        if require_group_id and int(require_group_id) != int(session['group_id']):
            _log.error("Attempted to install for group %s when group %s is only allowed" %
                       (session['group_id'], require_group_id))
            return _invalid_install("Only group %s is allowed to install this add-on" % require_group_id)

        client.group_id = session['group_id']
        client.group_name = session['group_name']
        clients.remove(client.id_query)
        clients.insert(client.to_map())
        if send_events:
            events.fire_event('install', {"client": client})
        return '', 201

    # noinspection PyUnusedLocal
    @addon.app.route('/addon/installable/<string:oauth_id>', methods=['DELETE'])
    def on_uninstall(oauth_id):
        uninstall_client(oauth_id, db_name, send_events)
        return '', 204

    addon.descriptor['capabilities']['installable']['callbackUrl'] = "{base}/addon/installable".format(
        base=addon.app.config['BASE_URL']
    )


def uninstall_client(oauth_id, db_name='clients', send_events=True):
    client = Tenant.load(oauth_id)
    clients = mongo[db_name]
    client_filter = {"id": oauth_id}
    clients.remove(client_filter)
    if send_events:
        events.fire_event('uninstall', {"client": client})


class Glance(object):

    def __init__(self):
        self.data = {}

    def with_label(self, value, glance_type="html"):
        self.data["label"] = {
            "value": value,
            "type": glance_type
        }
        return self

    def with_lozenge(self, label, lozenge_type):
        self.data["status"] = {
            "type": "lozenge",
            "value": {
                "label": label,
                "type": lozenge_type
            }
        }
        return self

    def with_icon(self, url, url2x):
        self.data["status"] = {
            "type": "icon",
            "value": {
                "url": url,
                "url@2x": url2x
            }
        }
        return self


import logging

_log = logging.getLogger(__name__)


class EventBus(object):

    def __init__(self):
        super(EventBus, self).__init__()
        self.events = {}

    def fire_event(self, name, obj):
        listeners = self.events.get(name, [])
        _log.debug("Firing event %s for %s listeners" % (name, len(listeners)))
        for listener in listeners:
            listener(obj)

    def register_event(self, name, func):
        _log.debug("Registering event: " + name)
        self.events.setdefault(name, []).append(func)

    def unregister_event(self, name, func):
        del self.events.setdefault(name, [])[func]

    def event_listener(self, func):
        self.register_event(func.__name__, func)
        return func

events = EventBus()


from datetime import timedelta
import time
from ac_flask.hipchat.db import mongo
import jwt
import logging
import requests
from requests.auth import HTTPBasicAuth
from werkzeug.exceptions import abort
from urlparse import urlparse

_log = logging.getLogger(__name__)

ACCESS_TOKEN_CACHE = "hipchat-tokens:{oauth_id}"


def base_url(url):
    if not url:
        return None
    result = urlparse(url)
    return "{scheme}://{netloc}".format(scheme=result.scheme, netloc=result.netloc)


class Tenant:

    def __init__(self, id, secret=None, homepage=None, capabilities_url=None, room_id=None, token_url=None,
                 group_id=None, group_name=None, capdoc=None):
        self.id = id
        self.room_id = room_id
        self.secret = secret
        self.group_id = group_id
        self.group_name = None if not group_name else group_name
        self.homepage = homepage or None if not capdoc else capdoc['links']['homepage']
        self.token_url = token_url or None if not capdoc else capdoc['capabilities']['oauth2Provider']['tokenUrl']
        self.capabilities_url = capabilities_url or None if not capdoc else capdoc['links']['self']
        self.api_base_url = capdoc['capabilities']['hipchatApiProvider']['url'] if capdoc \
            else self.capabilities_url[0:self.capabilities_url.rfind('/')] if self.capabilities_url else None
        self.installed_from = base_url(self.token_url)

    def to_map(self):
        return {
            "id": self.id,
            "secret": self.secret,
            "room_id": self.room_id,
            "group_id": self.group_id,
            "group_name": self.group_name,
            "homepage": self.homepage,
            "token_url": self.token_url,
            "capabilities_url": self.capabilities_url
        }

    @staticmethod
    def from_map(data):
        filtered = {key: val for key, val in data.items() if not key.startswith('_')}
        return Tenant(**filtered)

    @staticmethod
    def load(client_id):
        client_data = mongo.clients.find_one(Tenant(client_id).id_query)
        if client_data:
            return Tenant.from_map(client_data)
        else:
            _log.warn("Cannot find client: %s" % client_id)
            abort(400)

    @property
    def id_query(self):
        return {"id": self.id}

    def get_token(self, cache, token_only=True, scopes=None):
        if scopes is None:
            scopes = ["send_notification"]

        cache_key = ACCESS_TOKEN_CACHE.format(oauth_id=self.id)
        cache_key += ":" + ",".join(scopes)

        def gen_token():
            resp = requests.post(self.token_url, data={"grant_type": "client_credentials", "scope": " ".join(scopes)},
                                 auth=HTTPBasicAuth(self.id, self.secret), timeout=10)
            if resp.status_code == 200:
                _log.debug("Token request response: " + resp.text)
                return resp.json()
            elif resp.status_code == 401:
                _log.error("Client %s is invalid but we weren't notified.  Uninstalling" % self.id)
                raise OauthClientInvalidError(self)
            else:
                raise Exception("Invalid token: %s" % resp.text)

        if token_only:
            token = cache.get(cache_key)
            if not token:
                data = gen_token()
                token = data['access_token']
                cache.setex(cache_key, token, data['expires_in'] - 20)
            return token
        else:
            return gen_token()

    def sign_jwt(self, user_id, data=None):
        if data is None:
            data = {}

        now = int(time.time())
        exp = now + timedelta(hours=1).total_seconds()

        jwt_data = {"iss": self.id,
                    "iat": now,
                    "exp": exp}

        if user_id:
            jwt_data['sub'] = user_id

        data.update(jwt_data)
        return jwt.encode(data, self.secret)


class OauthClientInvalidError(Exception):
    def __init__(self, client, *args, **kwargs):
        super(OauthClientInvalidError, self).__init__(*args, **kwargs)
        self.client = client


import json
from ac_flask.hipchat.auth import tenant
from ac_flask.hipchat.db import redis
import requests


def post(url, data, token):
    return requests.post(url, headers={
        "authorization": "Bearer %s" % token,
        "content-type": "application/json"
    }, data=json.dumps(data), timeout=10)


class RoomClient(object):
    @staticmethod
    def send_notification(message):
        token = tenant.get_token(redis)
        return post("%s/room/%s/notification" % (tenant.api_base_url, tenant.room_id), {"message": message}, token)


room_client = RoomClient()


class AddOnClient(object):
    @staticmethod
    def update_global_glance(glance_key, glance_data):
        token = tenant.get_token(redis, scopes=['view_group'])
        return post("%s/addon/ui" % tenant.api_base_url, {
            "glance": [
                {"content": glance_data, "key": glance_key}
            ]
        }, token)

    @staticmethod
    def update_room_glance(glance_key, glance_data, room_id):
        token = tenant.get_token(redis, scopes=['view_room'])
        return post("%s/addon/ui/room/%s" % (tenant.api_base_url, room_id), {
            "glance": [
                {"content": glance_data, "key": glance_key}
            ]
        }, token)

    @staticmethod
    def update_user_glance(glance_key, glance_data, user_id):
        token = tenant.get_token(redis, scopes=['view_group'])
        return post("%s/addon/ui/user/%s" % (tenant.api_base_url, user_id), {
            "glance": [
                {"content": glance_data, "key": glance_key}
            ]
        }, token)


addon_client = AddOnClient()


"""
AC-Flask-HipChat
-------------

A library to help write a Flask-based HipChat add-on
"""
from setuptools import setup


setup(
    name='AC-Flask-HipChat',
    version='0.2.7',
    url='https://bitbucket.org/atlassianlabs/ac-flask-hipchat',
    license='APLv2',
    author='Don Brown',
    author_email='mrdon@twdata.org',
    description='Atlassian Connect library based on Flask for HipChat',
    long_description=__doc__,
    packages=['ac_flask', 'ac_flask.hipchat'],
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    install_requires=[
        'Flask',
        'pymongo',
        'redis',
        'requests',
        'PyJWT'
    ],
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)



from ac_flask.hipchat.tenant import Tenant
from flask import _request_ctx_stack as stack, request
from flask import abort
import jwt
from werkzeug.local import LocalProxy
from functools import wraps


def require_tenant(func):

    @wraps(func)
    def inner(*args, **kwargs):
        if not tenant:
            abort(401)
        return func(*args, **kwargs)

    return inner


def _validate_jwt(req):
    jwt_data = req.args.get('signed_request', None)

    if not jwt_data:
        header = req.headers.get('authorization', '')
        jwt_data = header[4:] if header.startswith('JWT ') else None

    if not jwt_data:
        abort(401)

    try:
        oauth_id = jwt.decode(jwt_data, verify=False)['iss']
        client = Tenant.load(oauth_id)
        data = jwt.decode(jwt_data, client.secret)
        return client, data

    except jwt.exceptions.DecodeError:
        abort(400)


def _get_tenant():
    ctx = stack.top
    if ctx is not None:
        if not hasattr(ctx, 'tenant'):
            body = request.json
            cur_sender = cur_context = None
            if request.args.get('signed_request', None) or 'authorization' in request.headers:
                cur_tenant, data = _validate_jwt(request)
                cur_sender = User(data['sub'])
                cur_context = data.get('context', None)
            elif body and 'oauth_client_id' in body:
                tenant_id = body['oauth_client_id']
                cur_tenant = Tenant.load(tenant_id)
            else:
                cur_tenant = None

            if body and 'item' in body:
                sent_by = _extract_sender(body['item'])
                if sent_by:
                    user = User(user_id=sent_by['id'], name=sent_by['name'], mention_name=sent_by['mention_name'])
                    # Check if the sender in the webhook matches the one provided in the JWT
                    if cur_sender and str(cur_sender.id) != str(user.id):
                        abort(400)
                    cur_sender = user

            ctx.tenant = cur_tenant
            ctx.sender = cur_sender
            ctx.context = cur_context

        return ctx.tenant


def _extract_sender(item):
    if 'sender' in item:
        return item['sender']
    if 'message' in item and 'from' in item['message']:
        return item['message']['from']
    return None


def _get_sender():
    _get_tenant()
    if hasattr(stack.top, 'sender'):
        return stack.top.sender
    else:
        return None


def _get_context():
    _get_tenant()
    if hasattr(stack.top, 'context'):
        return stack.top.context
    else:
        return None


tenant = LocalProxy(_get_tenant)
sender = LocalProxy(_get_sender)
context = LocalProxy(_get_context)


class User(object):
    def __init__(self, user_id, name=None, mention_name=None):
        super(User, self).__init__()
        self.id = user_id
        self.name = name
        self.mention_name = mention_name



from functools import wraps
import httplib
import logging

from ac_flask.hipchat import installable
from ac_flask.hipchat.auth import require_tenant, tenant
import os
from flask import jsonify, request
from urlparse import urlparse


_log = logging.getLogger(__name__)


def _not_none(app, name, default):
    val = app.config.get(name, default)
    if val is not None:
        return val
    else:
        raise ValueError("Missing '{key}' configuration property".format(key=name))


class Addon(object):

    def __init__(self, app, key=None, name=None, description=None, config=None, env_prefix="AC_",
                 allow_room=True, allow_global=False, scopes=None, vendor_name=None, vendor_url=None):
        if scopes is None:
            scopes = ['send_notification']

        self.app = app
        self._init_app(app, config, env_prefix)

        self.descriptor = {
            "key": _not_none(app, 'ADDON_KEY', key),
            "name": _not_none(app, 'ADDON_NAME', name),
            "description": app.config.get('ADDON_DESCRIPTION', description) or "",
            "links": {
                "self": self._relative_to_base("/addon/descriptor")
            },
            "capabilities": {
                "installable": {
                    "allowRoom": allow_room,
                    "allowGlobal": allow_global
                },
                "hipchatApiConsumer": {
                    "scopes": scopes
                }
            },
            "vendor": {
                "url": app.config.get('ADDON_VENDOR_URL', vendor_url) or "",
                "name": app.config.get('ADDON_VENDOR_NAME', vendor_name) or ""
            }
        }

        installable.init(addon=self,
                         allow_global=allow_global,
                         allow_room=allow_room)

        @self.app.route("/addon/descriptor")
        def descriptor():
            return jsonify(self.descriptor)

        self.app.route("/")(descriptor)

    @staticmethod
    def _init_app(app, config, env_prefix):
        app.config.from_object('ac_flask.hipchat.default_settings')
        if config is not None:
            app.config.from_object(config)

        if env_prefix is not None:
            env_vars = {key[len(env_prefix):]: val for key, val in os.environ.items()}
            app.config.update(env_vars)

        if app.config['DEBUG']:
            # These two lines enable debugging at httplib level (requests->urllib3->httplib)
            # You will see the REQUEST, including HEADERS and DATA, and RESPONSE with HEADERS but without DATA.
            # The only thing missing will be the response.body which is not logged.
            httplib.HTTPConnection.debuglevel = 1

            # You must initialize logging, otherwise you'll not see debug output.
            logging.basicConfig()
            logging.getLogger().setLevel(logging.DEBUG)
            requests_log = logging.getLogger("requests.packages.urllib3")
            requests_log.setLevel(logging.DEBUG)
            requests_log.propagate = True
        else:
            logging.basicConfig()
            logging.getLogger().setLevel(logging.WARN)

        app.events = {}

    def configure_page(self, path="/configure", **kwargs):
        self.descriptor['capabilities'].setdefault('configurable', {})['url'] = self._relative_to_base(path)

        def inner(func):
            return self.app.route(rule=path, **kwargs)(require_tenant(func))

        return inner

    def webhook(self, event, name=None, pattern=None, path=None, auth="jwt", **kwargs):
        if path is None:
            path = "/event/" + event

        wh = {
            "event": event,
            "url": self._relative_to_base(path),
            "authentication": auth
        }
        if name is not None:
            wh['name'] = name

        if pattern is not None:
            wh['pattern'] = pattern
        self.descriptor['capabilities'].setdefault('webhook', []).append(wh)

        def inner(func):
            return self.app.route(rule=path, methods=['POST'], **kwargs)(require_tenant(func))

        return inner

    def route(self, anonymous=False, *args, **kwargs):
        """
        Decorator for routes with defaulted required authenticated tenants
        """
        def inner(func):
            if not anonymous:
                func = require_tenant(func)
            func = self.app.route(*args, **kwargs)(func)
            return func

        return inner

    def glance(self, key, name, target, icon, icon2x=None, conditions=None, anonymous=False, path=None, **kwargs):

        if path is None:
            path = "/glance/" + key

        if icon2x is None:
            icon2x = icon

        glance_capability = {
            "key": key,
            "name": {
                "value": name
            },
            "queryUrl": self._relative_to_base(path),
            "target": target,
            "icon": {
                "url": self._relative_to_base(icon),
                "url@2x": self._relative_to_base(icon2x)
            },
            "conditions": conditions or []
        }

        self.descriptor['capabilities'].setdefault('glance', []).append(glance_capability)

        def inner(func):
            return self.route(anonymous, rule=path, **kwargs)(self.cors(self.json_output(func)))

        return inner

    def webpanel(self, key, name, location="hipchat.sidebar.right", anonymous=False, path=None, **kwargs):

        if path is None:
            path = "/webpanel/" + key

        webpanel_capability = {
            "key": key,
            "name": {
                "value": name
            },
            "url": self._relative_to_base(path),
            "location": location
        }

        self.descriptor['capabilities'].setdefault('webPanel', []).append(webpanel_capability)

        def inner(func):
            return self.route(anonymous, rule=path, **kwargs)(func)

        return inner

    def cors(self, func):
        @wraps(func)
        def inner(*args, **kwargs):
            whitelisted_origin = self._get_white_listed_origin()
            installed_from = tenant.installed_from if tenant else None
            response = self.app.make_response(func(*args, **kwargs))
            response.headers['Access-Control-Allow-Origin'] = whitelisted_origin or installed_from or '*'
            return response
        return inner

    def json_output(self, func):
        @wraps(func)
        def inner(*args, **kwargs):
            res = func(*args, **kwargs)
            return jsonify(res) if isinstance(res, dict) else res
        return inner

    def _relative_to_base(self, path):
        base = self.app.config['BASE_URL']
        path = '/' + path if not path.startswith('/') else path
        return base + path

    def _get_white_listed_origin(self):
        try:
            origin = request.headers['origin']
            if origin:
                origin_url = urlparse(origin)
                if origin_url and origin_url.hostname.endswith(self.app.config['CORS_WHITELIST']):
                    return origin
            return None
        except KeyError:
            return None

    def run(self, *args, **kwargs):
        if os.environ.get('WERKZEUG_RUN_MAIN') != 'true':
            print("")
            print("--------------------------------------")
            print("Public descriptor base URL: %s" % self.app.config['BASE_URL'])
            print("--------------------------------------")
            print("")

        self.app.run(*args, **kwargs)
